# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase


class TestIOCore(BlokTestCase):

    def checkExist(self):
        Mapping = self.registry.IO.Mapping
        self.assertTrue(
            Mapping.query().filter_by(
                key='test', model='Model.System.Blok').count())

    def checkUnExist(self):
        Mapping = self.registry.IO.Mapping
        self.assertFalse(
            Mapping.query().filter_by(
                key='test', model='Model.System.Blok').count())

    def test_session_delete_obj_with_mapping(self):
        Mapping = self.registry.IO.Mapping
        Blok = self.registry.System.Blok
        blok = Blok.insert(name='Test', version='0.0.0')
        self.checkUnExist()
        Mapping.set('test', blok)
        self.checkExist()
        blok.delete()
        self.checkUnExist()

    def test_query_delete_obj_with_mapping(self):
        Mapping = self.registry.IO.Mapping
        Blok = self.registry.System.Blok
        blok = Blok.insert(name='Test', version='0.0.0')
        self.checkUnExist()
        Mapping.set('test', blok)
        self.checkExist()
        Blok.query().filter_by(name='Test').delete()
        self.checkUnExist()



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .plugins import ModelPluginBase
from anyblok.mapper import ModelAttribute, ModelMapper


class ORMEventException(Exception):
    pass


class EventPlugin(ModelPluginBase):

    def __init__(self, registry):
        if not hasattr(registry, 'events'):
            registry.events = {}

        super(EventPlugin, self).__init__(registry)

    def transform_base_attribute(self, attr, method, namespace, base,
                                 transformation_properties,
                                 new_type_properties):
        """ Find the event listener methods in the base to save the
        namespace and the method in the registry

        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        if not hasattr(method, 'is_an_event_listener'):
            return
        elif method.is_an_event_listener is True:
            model = method.model
            event = method.event
            events = self.registry.events
            if model not in events:
                events[model] = {event: []}
            elif event not in events[model]:
                events[model][event] = []

            val = (namespace, attr)
            ev = events[model][event]
            if val not in ev:
                ev.append(val)


class SQLAlchemyEventPlugin(ModelPluginBase):

    def transform_base_attribute(self, attr, method, namespace, base,
                                 transformation_properties,
                                 new_type_properties):
        """declare in the registry the sqlalchemy event

        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        if not hasattr(method, 'is_an_sqlalchemy_event_listener'):
            return
        elif method.is_an_sqlalchemy_event_listener is True:
            self.registry._sqlalchemy_known_events.append(
                (method.sqlalchemy_listener,
                 namespace,
                 ModelAttribute(namespace, attr)))


class AutoSQLAlchemyORMEventPlugin(ModelPluginBase):

    def after_model_construction(self, base, namespace,
                                 transformation_properties):
        for eventtype in ('before_insert', 'after_insert',
                          'before_update', 'after_update',
                          'before_delete', 'after_delete'):
            attr = eventtype + '_orm_event'
            if hasattr(base, attr):
                if not hasattr(getattr(base, attr), '__self__'):
                    raise ORMEventException(
                        "On %s %s is not a classmethod" % (base, attr))

                self.registry._sqlalchemy_known_events.append((
                    ModelMapper(base, eventtype), namespace,
                    ModelAttribute(namespace, attr)))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok.column import Integer, String, Selection
from anyblok.relationship import Many2One, One2One, Many2Many, One2Many
from anyblok.declarations import Declarations
from anyblok.bloks.anyblok_core.exceptions import SqlBaseException


Model = Declarations.Model
register = Declarations.register


class TestCoreSQLBase(DBTestCase):

    def declare_model(self):
        from anyblok import Declarations
        Model = Declarations.Model

        @Declarations.register(Model)
        class Test:
            id = Integer(primary_key=True)
            id2 = Integer()

    def test_insert_and_query(self):
        registry = self.init_registry(self.declare_model)
        t1 = registry.Test.insert(id2=1)
        self.assertEqual(registry.Test.query().first(), t1)

    def test_multi_insert(self):
        registry = self.init_registry(self.declare_model)
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        self.assertEqual(registry.Test.query().count(), nb_value)
        for x in range(nb_value):
            self.assertEqual(
                registry.Test.query().filter(
                    registry.Test.id2 == x).count(),
                1)

    def test_delete(self):
        registry = self.init_registry(self.declare_model)
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        self.assertEqual(registry.Test.query().count(), nb_value)
        t = registry.Test.query().first()
        id2 = t.id2
        t.delete()
        self.assertEqual(registry.Test.query().count(), nb_value - 1)
        self.assertEqual(
            registry.Test.query().filter(
                registry.Test.id2 != id2).count(),
            nb_value - 1)

    def test_delete_with_get(self):
        registry = self.init_registry(self.declare_model)
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        self.assertEqual(registry.Test.query().count(), nb_value)
        t = registry.Test.query().first()
        self.assertIs(registry.Test.query().get(t.id), t)
        t.delete()
        self.assertIsNone(registry.Test.query().get(t.id))

    def test_expire(self):
        registry = self.init_registry(self.declare_model)
        t = registry.Test.insert(id2=2)
        self.assertEqual(t.id2, 2)
        t.id2 = 3
        self.assertEqual(t.id2, 3)
        t.expire()
        self.assertEqual(t.id2, 2)

    def test_refresh(self):
        registry = self.init_registry(self.declare_model)
        t = registry.Test.insert(id2=2)
        self.assertEqual(t.id2, 2)
        t.id2 = 3
        self.assertEqual(t.id2, 3)
        t.refresh()
        self.assertEqual(t.id2, 2)

    def test_delete_entry_added_in_relationship(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(len(t1.test2), 1)
        t2.delete()
        self.assertEqual(len(t1.test2), 0)

    def test_update(self):
        registry = self.init_registry(self.declare_model)
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = registry.Test.query().first()
        t.update(id2=100)
        self.assertEqual(
            registry.Test.query().filter(
                registry.Test.id2 == 100).first(),
            t)

    def test_get_primary_keys(self):
        registry = self.init_registry(self.declare_model)
        self.assertEqual(registry.Test.get_primary_keys(), ['id'])

    def test_to_and_from_primary_keys(self):
        registry = self.init_registry(self.declare_model)
        nb_value = 3
        registry.Test.multi_insert(*[{'id2': x} for x in range(nb_value)])
        t = registry.Test.query().first()
        self.assertEqual(t.to_primary_keys(), {'id': t.id})
        self.assertEqual(registry.Test.from_primary_keys(id=t.id), t)

    def add_in_registry_m2o(self):

        @register(Model)
        class Test:
            id = Integer(primary_key=True)
            name = String()

        @register(Model)
        class Test2:
            id = Integer(primary_key=True)
            name = String()
            test = Many2One(model=Model.Test, one2many="test2")

    def add_in_registry_o2m(self):

        @register(Model)
        class Test:
            id = Integer(primary_key=True)
            name = String()

        @register(Model)
        class Test2:
            id = Integer(primary_key=True)
            name = String()
            test_id = Integer(foreign_key=Model.Test.use('id'))

        @register(Model)  # noqa
        class Test:
            test2 = One2Many(model=Model.Test2, many2one="test")

    def add_in_registry_o2o(self):

        @register(Model)
        class Test:
            id = Integer(primary_key=True)
            name = String()

        @register(Model)
        class Test2:
            id = Integer(primary_key=True)
            name = String()
            test = One2One(model=Model.Test, backref='test2')

    def add_in_registry_m2m(self):

        @register(Model)
        class Test:
            id = Integer(primary_key=True)
            name = String()

        @register(Model)
        class Test2:
            id = Integer(primary_key=True)
            name = String()
            test = Many2Many(model=Model.Test, many2many='test2')

    def test_to_dict_m2o_with_pks(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t1.to_dict('name'), {'name': 't1'})
        self.assertEqual(t2.to_dict('name'), {'name': 't2'})
        self.assertEqual(t2.to_dict('name', 'test'),
                         {'name': 't2', 'test': {'id': t1.id}})

    def test_to_dict_o2o_with_pks(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t1.to_dict('name'), {'name': 't1'})
        self.assertEqual(t2.to_dict('name'), {'name': 't2'})
        self.assertEqual(t2.to_dict('name', 'test'),
                         {'name': 't2', 'test': {'id': t1.id}})
        self.assertEqual(t1.to_dict('name', 'test2'),
                         {'name': 't1', 'test2': {'id': t2.id}})

    def test_to_dict_m2m_with_pks(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        self.assertEqual(t1.to_dict('name'), {'name': 't1'})
        self.assertEqual(t2.to_dict('name'), {'name': 't2'})
        self.assertEqual(t2.to_dict('name', 'test'),
                         {'name': 't2', 'test': [{'id': t1.id}]})
        self.assertEqual(t1.to_dict('name', 'test2'),
                         {'name': 't1', 'test2': [{'id': t2.id}]})

    def test_to_dict_o2m_with_pks(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t1.to_dict('name', 'test2'),
                         {'name': 't1', 'test2': [{'id': t2.id}]})

    def test_to_dict_m2o_with_column(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t2.to_dict('name', ('test', ('name',))),
                         {'name': 't2', 'test': {'name': 't1'}})

    def test_to_dict_o2o_with_column(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t2.to_dict('name', ('test', ('name',))),
                         {'name': 't2', 'test': {'name': 't1'}})
        self.assertEqual(t1.to_dict('name', ('test2', ('name',))),
                         {'name': 't1', 'test2': {'name': 't2'}})

    def test_to_dict_m2m_with_column(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        self.assertEqual(t2.to_dict('name', ('test', ('name',))),
                         {'name': 't2', 'test': [{'name': 't1'}]})
        self.assertEqual(t1.to_dict('name', ('test2', ('name',))),
                         {'name': 't1', 'test2': [{'name': 't2'}]})

    def test_to_dict_o2m_with_column(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t1.to_dict('name', ('test2', ('name',))),
                         {'name': 't1', 'test2': [{'name': 't2'}]})

    def test_to_dict_m2o_with_all_columns(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t2.to_dict('name', ('test',)),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': [{'id': t2.id}]}})
        self.assertEqual(t2.to_dict('name', ('test', None)),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': [{'id': t2.id}]}})
        self.assertEqual(t2.to_dict('name', ('test', ())),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': [{'id': t2.id}]}})

    def test_to_dict_o2o_with_all_columns(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t2.to_dict('name', ('test',)),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': {'id': t2.id}}})
        self.assertEqual(t2.to_dict('name', ('test', None)),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': {'id': t2.id}}})
        self.assertEqual(t2.to_dict('name', ('test', ())),
                         {'name': 't2', 'test': {'name': 't1',
                                                 'id': t1.id,
                                                 'test2': {'id': t2.id}}})

    def test_to_dict_m2m_with_all_columns(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        self.assertEqual(t2.to_dict('name', ('test',)),
                         {'name': 't2', 'test': [{'name': 't1',
                                                  'id': t1.id,
                                                  'test2': [{'id': t2.id}]}]})
        self.assertEqual(t2.to_dict('name', ('test', None)),
                         {'name': 't2', 'test': [{'name': 't1',
                                                  'id': t1.id,
                                                  'test2': [{'id': t2.id}]}]})
        self.assertEqual(t2.to_dict('name', ('test', ())),
                         {'name': 't2', 'test': [{'name': 't1',
                                                  'id': t1.id,
                                                  'test2': [{'id': t2.id}]}]})
        self.assertEqual(t1.to_dict('name', ('test2',)),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test': [{'id': t1.id}]}]})
        self.assertEqual(t1.to_dict('name', ('test2', None)),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test': [{'id': t1.id}]}]})
        self.assertEqual(t1.to_dict('name', ('test2', ())),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test': [{'id': t1.id}]}]})

    def test_to_dict_o2m_with_all_columns(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(t1.to_dict('name', ('test2',)),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test_id': t1.id,
                                                   'test': {'id': t1.id}}]})
        self.assertEqual(t1.to_dict('name', ('test2', None)),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test_id': t1.id,
                                                   'test': {'id': t1.id}}]})
        self.assertEqual(t1.to_dict('name', ('test2', ())),
                         {'name': 't1', 'test2': [{'name': 't2',
                                                   'id': t2.id,
                                                   'test_id': t1.id,
                                                   'test': {'id': t1.id}}]})

    def test_bad_definition_of_relation(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with self.assertRaises(SqlBaseException):
            t2.to_dict('name', ('test', 'name'))

    def test_bad_definition_of_relation2(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with self.assertRaises(SqlBaseException):
            t2.to_dict('name', ('test', ('name',), ()))

    def test_bad_definition_of_relation3(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        with self.assertRaises(SqlBaseException):
            t2.to_dict('name', ())

    def test_refresh_update_m2o(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertEqual(t2.test_id, t1.id)
        self.assertEqual(t3.test2, [])
        t2.test = t3
        self.assertEqual(t1.test2, [])
        self.assertIs(t2.test, t3)
        self.assertIs(t2.test_id, t3.id)
        self.assertEqual(t3.test2, [t2])
        t2.test = None
        self.assertEqual(t1.test2, [])
        self.assertIsNone(t2.test)
        self.assertIsNone(t2.test_id)
        self.assertEqual(t3.test2, [])

    def test_refresh_update_m2o_2(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertEqual(t2.test_id, t1.id)
        self.assertEqual(t3.test2, [])
        t2.test_id = t3.id
        self.assertEqual(t1.test2, [])
        self.assertIs(t2.test, t3)
        self.assertEqual(t2.test_id, t3.id)
        self.assertEqual(t3.test2, [t2])

    def test_refresh_update_m2o_3(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertEqual(t2.test_id, t1.id)
        self.assertEqual(t3.test2, [])
        t3.test2.append(t2)
        self.assertEqual(t1.test2, [])
        self.assertIs(t2.test, t3)
        self.assertEqual(t2.test_id, t3.id)
        self.assertEqual(t3.test2, [t2])
        t3.test2.remove(t2)
        self.assertEqual(t1.test2, [])
        self.assertIsNone(t2.test)
        self.assertIsNone(t2.test_id)
        self.assertEqual(t3.test2, [])

    def test_refresh_update_o2o(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertIs(t1.test2, t2)
        self.assertIs(t2.test, t1)
        self.assertIs(t2.test_id, t1.id)
        self.assertIsNone(t3.test2)
        t2.test = t3
        self.assertIsNone(t1.test2)
        self.assertIs(t2.test, t3)
        self.assertEqual(t2.test_id, t3.id)
        self.assertIs(t3.test2, t2)

    def test_refresh_update_o2o_2(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertIs(t1.test2, t2)
        self.assertIs(t2.test, t1)
        self.assertIsNone(t3.test2)
        t3.test2 = t2
        self.assertIsNone(t1.test2)
        self.assertIs(t2.test, t3)
        self.assertIs(t2.test_id, t3.id)
        self.assertIs(t3.test2, t2)

    def test_refresh_update_o2o_3(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test.insert(name='t3')
        self.assertIs(t1.test2, t2)
        self.assertIs(t2.test, t1)
        self.assertEqual(t2.test_id, t1.id)
        self.assertIsNone(t3.test2)
        t2.test_id = t3.id
        self.assertIsNone(t1.test2)
        self.assertIs(t2.test, t3)
        self.assertEqual(t2.test_id, t3.id)
        self.assertIs(t3.test2, t2)

    def test_refresh_update_o2m(self):
        registry = self.init_registry(self.add_in_registry_o2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertIsNone(t3.test)
        t1.test2.append(t3)
        self.assertIn(t2, t1.test2)
        self.assertIn(t3, t1.test2)
        self.assertIs(t2.test, t1)
        self.assertIs(t3.test, t1)
        t1.test2.remove(t2)
        self.assertEqual(t1.test2, [t3])
        self.assertIsNone(t2.test)
        self.assertIs(t3.test, t1)

    def test_refresh_update_o2m_2(self):
        registry = self.init_registry(self.add_in_registry_o2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertIsNone(t3.test)
        t3.test = t1
        self.assertIn(t2, t1.test2)
        self.assertIn(t3, t1.test2)
        self.assertIs(t2.test, t1)
        self.assertIs(t3.test, t1)
        self.assertIs(t3.test_id, t1.id)

    def test_refresh_update_o2m_3(self):
        registry = self.init_registry(self.add_in_registry_o2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        t3 = registry.Test2.insert(name='t3')
        self.assertEqual(t1.test2, [t2])
        self.assertIs(t2.test, t1)
        self.assertIsNone(t3.test)
        t3.test_id = t1.id
        self.assertIn(t2, t1.test2)
        self.assertIn(t3, t1.test2)
        self.assertIs(t2.test, t1)
        self.assertIs(t3.test, t1)

    def test_refresh_update_m2m(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t3 = registry.Test2.insert(name='t3')
        self.assertEqual(t1.test2, [])
        self.assertEqual(t2.test, [])
        self.assertEqual(t3.test, [])
        t1.test2.append(t2)
        self.assertEqual(t1.test2, [t2])
        self.assertEqual(t2.test, [t1])
        self.assertEqual(t3.test, [])
        t1.test2.append(t3)
        self.assertIn(t2, t1.test2)
        self.assertIn(t3, t1.test2)
        self.assertEqual(t2.test, [t1])
        self.assertEqual(t3.test, [t1])
        t1.test2.remove(t2)
        self.assertEqual(t1.test2, [t3])
        self.assertEqual(t2.test, [])
        self.assertEqual(t3.test, [t1])

    def test_refresh_update_m2m_2(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t3 = registry.Test2.insert(name='t3')
        self.assertEqual(t1.test2, [])
        self.assertEqual(t2.test, [])
        self.assertEqual(t3.test, [])
        t2.test.append(t1)
        self.assertEqual(t1.test2, [t2])
        self.assertEqual(t2.test, [t1])
        self.assertEqual(t3.test, [])
        t3.test.append(t1)
        self.assertIn(t2, t1.test2)
        self.assertIn(t3, t1.test2)
        self.assertEqual(t2.test, [t1])
        self.assertEqual(t3.test, [t1])
        t2.test.remove(t1)
        self.assertEqual(t1.test2, [t3])
        self.assertEqual(t2.test, [])
        self.assertEqual(t3.test, [t1])

    def test_find_relationship_by_relationship(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        fields = registry.Test2.find_relationship('test')
        self.assertIn('test', fields)
        self.assertIn('test_id', fields)

    def test_find_relationship_by_column(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        fields = registry.Test2.find_relationship('test_id')
        self.assertIn('test', fields)
        self.assertIn('test_id', fields)

    def declare_model_with_column_selection(self):
        from anyblok import Declarations
        Model = Declarations.Model

        @Declarations.register(Model)
        class Test:
            id = Integer(primary_key=True)
            select = Selection(
                selections=[('key', 'value'), ('key2', 'value2')],
                default='key')

    def test_expire_with_column_selection(self):
        registry = self.init_registry(self.declare_model_with_column_selection)
        t = registry.Test.insert()
        t.select = 'key2'
        t.expire('select')
        self.assertEqual(t.select, 'key')

    def test_getFieldType(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        self.assertEqual(registry.Test.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Test.getFieldType('name'), 'String')
        self.assertEqual(registry.Test2.getFieldType('id'), 'Integer')
        self.assertEqual(registry.Test2.getFieldType('name'), 'String')
        self.assertEqual(registry.Test2.getFieldType('test'), 'Many2One')
        self.assertEqual(registry.Test.getFieldType('test2'), 'One2Many')

    def test_repr_m2o(self):
        registry = self.init_registry(self.add_in_registry_m2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(
            repr(t2),
            "<Model.Test2(id=1, name='t2', test=<Model.Test(id=1)>"
            ", test_id=1)>"
        )

    def test_repr_o2o(self):
        registry = self.init_registry(self.add_in_registry_o2o)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(
            repr(t2),
            "<Model.Test2(id=1, name='t2', test=<Model.Test(id=1)>"
            ", test_id=1)>"
        )

    def test_repr_o2m(self):
        registry = self.init_registry(self.add_in_registry_o2m)
        t1 = registry.Test.insert(name='t1')
        registry.Test2.insert(name='t2', test=t1)
        self.assertEqual(
            repr(t1),
            "<Model.Test(id=1, name='t1', test2=<not loaded>)>"
        )

    def test_repr_m2m(self):
        registry = self.init_registry(self.add_in_registry_m2m)
        t1 = registry.Test.insert(name='t1')
        t2 = registry.Test2.insert(name='t2')
        t2.test.append(t1)
        self.assertEqual(
            repr(t2),
            "<Model.Test2(id=1, name='t2', test=<Model.Test len(1)>)>"
        )



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from random import random
from anyblok.tests.testcase import DBTestCase
from anyblok.declarations import Declarations, cache, classmethod_cache
from anyblok.bloks.anyblok_core.exceptions import CacheException
from anyblok.column import Integer
register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin
Core = Declarations.Core


def wrap_cache(condition):

    def wrapper(function):
        return function

    if condition:
        return cache()

    return wrapper


def wrap_cls_cache(condition):
    if condition:
        return classmethod_cache()

    return classmethod


class TestCache(DBTestCase):

    def add_model_with_method_cached(self):

        @register(Model)
        class Test:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

    def test_cache_invalidation(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        Cache = registry.System.Cache
        nb_invalidation = Cache.query().count()
        Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(Cache.query().count(), nb_invalidation + 1)

    def test_invalid_cache_invalidation(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        Cache = registry.System.Cache
        with self.assertRaises(CacheException):
            Cache.invalidate('Model.Test2', 'method_cached')

    def test_detect_cache_invalidation(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        Cache = registry.System.Cache
        self.assertEqual(Cache.detect_invalidation(), False)
        Cache.insert(registry_name="Model.Test", method="method_cached")
        self.assertEqual(Cache.detect_invalidation(), True)

    def test_get_invalidation(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        Cache = registry.System.Cache
        Cache.insert(registry_name="Model.Test", method="method_cached")
        caches = Cache.get_invalidation()
        self.assertEqual(len(caches), 1)
        cache = caches[0]
        self.assertEqual(cache.indentify, ('Model.Test', 'method_cached'))


class TestSimpleCache(DBTestCase):

    def check_method_cached(self, Model, registry_name, value=1):
        m = Model()
        self.assertEqual(m.method_cached(), value)
        self.assertEqual(m.method_cached(), value)
        Model.registry.System.Cache.invalidate(registry_name, 'method_cached')
        self.assertEqual(m.method_cached(), 2 * value)

    def check_method_cached_invalidate_all(self, Model, value=1):
        m = Model()
        self.assertEqual(m.method_cached(), value)
        self.assertEqual(m.method_cached(), value)
        Model.registry.System.Cache.invalidate_all()
        self.assertEqual(m.method_cached(), 2 * value)

    def add_model_with_method_cached(self):

        @register(Model)
        class Test:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

    def add_model_with_method_cached_by_core(self):

        @register(Core)
        class Base:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Model)
        class Test:
            pass

    def add_model_with_method_cached_by_mixin(self):

        @register(Mixin)
        class MTest:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Model)
        class Test(Mixin.MTest):
            pass

    def add_model_with_method_cached_by_mixin_chain(self):

        @register(Mixin)
        class MTest:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Mixin)
        class MTest2(Mixin.MTest):
            pass

        @register(Model)
        class Test(Mixin.MTest2):
            pass

    def add_model_with_method_cached_with_mixin_and_or_core(self,
                                                            withmodel=False,
                                                            withmixin=False,
                                                            withcore=False):

        @register(Core)
        class Base:

            x = 0

            @wrap_cache(withcore)
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Mixin)
        class MTest:

            y = 0

            @wrap_cache(withmixin)
            def method_cached(self):
                self.y += 2
                return self.y + super(MTest, self).method_cached()

        @register(Model)
        class Test(Mixin.MTest):

            z = 0

            @wrap_cache(withmodel)
            def method_cached(self):
                self.z += 3
                return self.z + super(Test, self).method_cached()

    def test_model(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_model2(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_core(self):
        registry = self.init_registry(self.add_model_with_method_cached_by_core)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_core2(self):
        registry = self.init_registry(self.add_model_with_method_cached_by_core)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_mixin2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_mixin_chain(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin_chain)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_mixin_chain2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin_chain)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_model_mixin_core_not_cache(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core)
        m = registry.Test()
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 12)

    def test_model_mixin_core_only_core(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withcore=True)
        m = registry.Test()
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 11)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 17)

    def test_model_mixin_core_only_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmixin=True)
        m = registry.Test()
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 9)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 15)

    def test_model_mixin_core_only_model(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmodel=True)
        m = registry.Test()
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 6)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 12)

    def test_model_mixin_core_only_core_and_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmixin=True, withcore=True)
        m = registry.Test()
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 9)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 15)

    def test_invalidate_all_check_model(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        self.check_method_cached_invalidate_all(registry.Test)

    def test_invalidate_all_check_core(self):
        registry = self.init_registry(self.add_model_with_method_cached_by_core)
        self.check_method_cached_invalidate_all(registry.Test)

    def test_invalidate_all_check_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        self.check_method_cached_invalidate_all(registry.Test)

    def add_model_with_method_core_cached_with_two_model(self):

        @register(Core)
        class Base:

            def __init__(self):
                super(Base, self).__init__()
                self.x = 0
                self.z = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Model)
        class Test:

            def method_cached(self):
                self.z += 3
                return self.z + super(Test, self).method_cached()

        @register(Model)
        class Test2:

            def method_cached(self):
                self.z += 3
                return self.z + super(Test2, self).method_cached()

    def test_2_model_with_core_catched(self):
        registry = self.init_registry(
            self.add_model_with_method_core_cached_with_two_model)
        t = registry.Test()
        t2 = registry.Test2()
        self.assertEqual(t.method_cached(), 4)
        self.assertEqual(t2.method_cached(), 4)
        self.assertEqual(t.method_cached(), 7)
        self.assertEqual(t2.method_cached(), 7)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(t.method_cached(), 11)
        self.assertEqual(t2.method_cached(), 10)


class TestClassMethodCache(DBTestCase):

    def check_method_cached(self, Model, registry_name):
        m = Model
        value = m.method_cached()
        self.assertEqual(m.method_cached(), value)
        self.assertEqual(m.method_cached(), value)
        Model.registry.System.Cache.invalidate(registry_name, 'method_cached')
        self.assertNotEqual(m.method_cached(), value)

    def add_model_with_method_cached(self):

        @register(Model)
        class Test:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

    def add_model_with_method_cached_by_core(self):

        @register(Core)
        class Base:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Model)
        class Test:
            pass

    def add_model_with_method_cached_by_mixin(self):

        @register(Mixin)
        class MTest:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Model)
        class Test(Mixin.MTest):
            pass

    def add_model_with_method_cached_by_mixin_chain(self):

        @register(Mixin)
        class MTest:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Mixin)
        class MTest2(Mixin.MTest):
            pass

        @register(Model)
        class Test(Mixin.MTest2):
            pass

    def add_model_with_method_cached_with_mixin_and_or_core(self,
                                                            withmodel=False,
                                                            withmixin=False,
                                                            withcore=False):

        @register(Core)
        class Base:

            x = 0

            @wrap_cls_cache(withcore)
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Mixin)
        class MTest:

            y = 0

            @wrap_cls_cache(withmixin)
            def method_cached(cls):
                cls.y += 2
                return cls.y + super(MTest, cls).method_cached()

        @register(Model)
        class Test(Mixin.MTest):

            z = 0

            @wrap_cls_cache(withmodel)
            def method_cached(cls):
                cls.z += 3
                return cls.z + super(Test, cls).method_cached()

    def test_model(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_model2(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_core(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_core)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_core2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_core)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_mixin2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_mixin_chain(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin_chain)
        self.check_method_cached(registry.Test, 'Model.Test')

    def test_mixin_chain2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin_chain)
        from anyblok import Declarations
        self.check_method_cached(registry.Test, Declarations.Model.Test)

    def test_model_mixin_core_not_cache(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core)
        m = registry.Test
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 12)

    def test_model_mixin_core_only_core(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withcore=True)
        m = registry.Test
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 11)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 17)

    def test_model_mixin_core_only_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmixin=True)
        m = registry.Test
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 9)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 15)

    def test_model_mixin_core_only_model(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmodel=True)
        m = registry.Test
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 6)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 12)

    def test_model_mixin_core_only_core_and_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_with_mixin_and_or_core,
            withmixin=True, withcore=True)
        m = registry.Test
        self.assertEqual(m.method_cached(), 6)
        self.assertEqual(m.method_cached(), 9)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 15)

    def add_model_with_method_core_cached_with_two_model(self):

        @register(Core)
        class Base:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Model)
        class Test:

            z = 0

            @classmethod
            def method_cached(cls):
                cls.z += 3
                return cls.z + super(Test, cls).method_cached()

        @register(Model)
        class Test2:

            z = 0

            @classmethod
            def method_cached(cls):
                cls.z += 3
                return cls.z + super(Test2, cls).method_cached()

    def test_2_model_with_core_catched(self):
        registry = self.init_registry(
            self.add_model_with_method_core_cached_with_two_model)
        self.assertEqual(registry.Test.method_cached(), 4)
        self.assertEqual(registry.Test2.method_cached(), 4)
        self.assertEqual(registry.Test.method_cached(), 7)
        self.assertEqual(registry.Test2.method_cached(), 7)
        registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(registry.Test.method_cached(), 11)
        self.assertEqual(registry.Test2.method_cached(), 10)


class TestInheritedCache(DBTestCase):

    def check_method_cached(self, Model):
        m = Model()
        self.assertEqual(m.method_cached(), 3)
        self.assertEqual(m.method_cached(), 5)
        Model.registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 8)

    def check_inherited_method_cached(self, Model):
        m = Model()
        self.assertEqual(m.method_cached(), 3)
        self.assertEqual(m.method_cached(), 3)
        Model.registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(m.method_cached(), 6)

    def add_model_with_method_cached(self, inheritcache=False):

        @register(Model)
        class Test:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Model)  # noqa
        class Test:

            y = 0

            if inheritcache:
                @cache()
                def method_cached(self):
                    self.y += 2
                    return self.y + super(Test, self).method_cached()
            else:
                def method_cached(self):
                    self.y += 2
                    return self.y + super(Test, self).method_cached()

    def add_model_with_method_cached_by_core(self, inheritcache=False):

        @register(Core)
        class Base:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Core)  # noqa
        class Base:

            y = 0

            if inheritcache:
                @cache()
                def method_cached(self):
                    self.y += 2
                    return self.y + super(Base, self).method_cached()
            else:
                def method_cached(self):
                    self.y += 2
                    return self.y + super(Base, self).method_cached()

        @register(Model)
        class Test:
            pass

    def add_model_with_method_cached_by_mixin(self, inheritcache=False):

        @register(Mixin)
        class MTest:

            x = 0

            @cache()
            def method_cached(self):
                self.x += 1
                return self.x

        @register(Mixin)  # noqa
        class MTest:

            y = 0

            if inheritcache:
                @cache()
                def method_cached(self):
                    self.y += 2
                    return self.y + super(MTest, self).method_cached()
            else:
                def method_cached(self):
                    self.y += 2
                    return self.y + super(MTest, self).method_cached()

        @register(Model)
        class Test(Mixin.MTest):
            pass

    def test_model(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        self.check_method_cached(registry.Test)

    def test_model2(self):
        registry = self.init_registry(self.add_model_with_method_cached,
                                      inheritcache=True)
        self.check_inherited_method_cached(registry.Test)

    def test_core(self):
        registry = self.init_registry(self.add_model_with_method_cached_by_core)
        self.check_method_cached(registry.Test)

    def test_core2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_core, inheritcache=True)
        self.check_inherited_method_cached(registry.Test)

    def test_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        self.check_method_cached(registry.Test)

    def test_mixin2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin, inheritcache=True)
        self.check_inherited_method_cached(registry.Test)


class TestInheritedClassMethodCache(DBTestCase):

    def check_method_cached(self, Model):
        self.assertEqual(Model.method_cached(), 3)
        self.assertEqual(Model.method_cached(), 5)
        Model.registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(Model.method_cached(), 8)

    def check_inherited_method_cached(self, Model):
        self.assertEqual(Model.method_cached(), 3)
        self.assertEqual(Model.method_cached(), 3)
        Model.registry.System.Cache.invalidate('Model.Test', 'method_cached')
        self.assertEqual(Model.method_cached(), 6)

    def add_model_with_method_cached(self, inheritcache=False):

        @register(Model)
        class Test:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Model)  # noqa
        class Test:

            y = 0

            if inheritcache:
                @classmethod_cache()
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(Test, cls).method_cached()
            else:
                @classmethod
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(Test, cls).method_cached()

    def add_model_with_method_cached_by_core(self, inheritcache=False):

        @register(Core)
        class Base:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Core)  # noqa
        class Base:

            y = 0

            if inheritcache:
                @classmethod_cache()
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(Base, cls).method_cached()
            else:
                @classmethod
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(Base, cls).method_cached()

        @register(Model)
        class Test:
            pass

    def add_model_with_method_cached_by_mixin(self, inheritcache=False):

        @register(Mixin)
        class MTest:

            x = 0

            @classmethod_cache()
            def method_cached(cls):
                cls.x += 1
                return cls.x

        @register(Mixin)  # noqa
        class MTest:

            y = 0

            if inheritcache:
                @classmethod_cache()
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(MTest, cls).method_cached()
            else:
                @classmethod
                def method_cached(cls):
                    cls.y += 2
                    return cls.y + super(MTest, cls).method_cached()

        @register(Model)
        class Test(Mixin.MTest):
            pass

    def test_model(self):
        registry = self.init_registry(self.add_model_with_method_cached)
        self.check_method_cached(registry.Test)

    def test_model2(self):
        registry = self.init_registry(self.add_model_with_method_cached,
                                      inheritcache=True)
        self.check_inherited_method_cached(registry.Test)

    def test_core(self):
        registry = self.init_registry(self.add_model_with_method_cached_by_core)
        self.check_method_cached(registry.Test)

    def test_core2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_core, inheritcache=True)
        self.check_inherited_method_cached(registry.Test)

    def test_mixin(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin)
        self.check_method_cached(registry.Test)

    def test_mixin2(self):
        registry = self.init_registry(
            self.add_model_with_method_cached_by_mixin, inheritcache=True)
        self.check_inherited_method_cached(registry.Test)


class TestComparatorInterModel(DBTestCase):

    def check_comparator(self, registry):
        Test = registry.Test
        Test2 = registry.Test2
        self.assertEqual(Test.method_cached(), Test.method_cached())
        self.assertEqual(Test2.method_cached(), Test2.method_cached())
        self.assertNotEqual(Test.method_cached(), Test2.method_cached())

    def test_model(self):

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod_cache()
                def method_cached(cls):
                    return random()

            @register(Model)
            class Test2:

                @classmethod_cache()
                def method_cached(cls):
                    return random()

        registry = self.init_registry(add_in_registry)
        self.check_comparator(registry)

    def test_mixin(self):

        def add_in_registry():

            @register(Mixin)
            class MTest:

                @classmethod_cache()
                def method_cached(cls):
                    return random()

            @register(Model)
            class Test(Mixin.MTest):
                pass

            @register(Model)
            class Test2(Mixin.MTest):

                pass

        registry = self.init_registry(add_in_registry)
        self.check_comparator(registry)

    def test_core(self):

        def add_in_registry():

            @register(Core)
            class Base:

                @classmethod_cache()
                def method_cached(cls):
                    return random()

            @register(Model)
            class Test:
                pass

            @register(Model)
            class Test2:

                pass

        registry = self.init_registry(add_in_registry)
        self.check_comparator(registry)


class TestSQLModelCache(DBTestCase):

    def add_in_registry(self):

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            id2 = Integer()

            @cache()
            def get_id2(self):
                return self.id2

            @classmethod_cache()
            def count(cls):
                return cls.query().count()

    def test_query_and_classmethod_cached(self):
        registry = self.init_registry(self.add_in_registry)
        registry.Test.insert()
        registry.Test.insert()
        self.assertEqual(registry.Test.count(), 2)
        registry.Test.insert()
        self.assertEqual(registry.Test.count(), 2)
        Cache = registry.System.Cache
        Cache.invalidate('Model.Test', 'count')
        self.assertEqual(registry.Test.count(), 3)

    def test_query_and_method_cached(self):
        registry = self.init_registry(self.add_in_registry)
        t = registry.Test.insert(id2=1)
        self.assertEqual(t.get_id2(), 1)
        t.id2 = 2
        self.assertEqual(t.get_id2(), 1)
        Cache = registry.System.Cache
        Cache.invalidate('Model.Test', 'get_id2')
        self.assertEqual(t.get_id2(), 2)



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .logging import log
from argparse import ArgumentParser, _ArgumentGroup, REMAINDER
from configparser import ConfigParser
import sys
import os
import json
import yaml
from appdirs import AppDirs
from os.path import join, isfile
from sqlalchemy.engine.url import URL, make_url
from logging import (getLogger, config, NOTSET, DEBUG, INFO, WARNING, ERROR,
                     CRITICAL, basicConfig, Logger)
logger = getLogger(__name__)


def get_db_name():
    """ Return an sqlalchemy name of the database from configuration
    db_name or db_url

    :rtype: name of the database
    :exception: ConfigurationException
    """
    url = Configuration.get('db_url', None)
    db_name = Configuration.get('db_name', None)

    if db_name is not None:
        return db_name

    if url:
        url = make_url(url)
        if url.database:
            return url.database

    raise ConfigurationException("No database's name defined")


def get_url(db_name=None):
    """ Return an sqlalchemy URL for database

    Get the options of the database, the only option which can be
    overloaded is the name of the database::

        url = get_url(db_name='Mydb')

    ..note::

        Since 0.5.3, an URL can be define by the configuration file.
        The *username*, *password* and *database* if overwrite by the
        options if they are filled::

            # db_url = 'postgresql:///db'
            get_url()
            ==> 'postgresql:///db'
            # db_user_name = 'jssuzanne'
            # db_password = 'secret'
            get_url()
            ==> 'postgresql://jssuzanne:secret@/db'
            # db_name = 'db1'
            get_url()
            ==> 'postgresql://jssuzanne:secret@/db1'
            get_url(db_name='Mydb')
            ==> 'postgresql://jssuzanne:secret@/Mydb'

    :param db_name: Name of the database
    :rtype: SqlAlchemy URL
    :exception: ConfigurationException
    """
    url = Configuration.get('db_url', None)
    drivername = Configuration.get('db_driver_name', None)
    username = Configuration.get('db_user_name', None)
    password = Configuration.get('db_password', None)
    host = Configuration.get('db_host', None)
    port = Configuration.get('db_port', None)
    database = Configuration.get('db_name', None)

    if db_name is not None:
        database = db_name

    if url:
        url = make_url(url)
        if username:
            url.username = username
        if password:
            url.password = password
        if database:
            url.database = database

        return url

    if drivername is None:
        raise ConfigurationException('No Drivername defined')

    return URL(drivername, username=username, password=password, host=host,
               port=port, database=database)


class ConfigurationException(LookupError):
    """ Simple Exception for Configuration"""


def is_none(type, value):
    """Check if the value is a *NoneValue* in function of the type

    :param type: type of value
    :param value: value to check
    :rtype: *bool*, True if it is a *NoneValue*
    """
    if value is None:
        return True
    if isinstance(value, str) and value.upper() == 'NONE':
        return True
    if value == '' and type is not str:
        return True

    return False


def cast_value(cast, value):
    """Cast the value

    :param cast: final type of data wanted
    :param value: initial value to cast
    :rtype: casted value or *NoneValue*
    """
    if is_none(cast, value):
        return None
    elif not cast:
        return value
    elif cast is bool and isinstance(value, str):
        if value.upper() == 'TRUE':
            return True
        else:
            return False
    else:
        return cast(value)


def nargs_type(key, nargs, cast):

    def wrap(val):
        if isinstance(val, str):
            sep = ' '
            if '\n' in val:
                sep = '\n'
            elif ',' in val:
                sep = ','
            val = [x.strip() for x in val.split(sep)]

        if not isinstance(val, list):
            raise ConfigurationException("Waiting list not %r for %r "
                                         "get : %r" % (type(val), key, val))

        limit = len(val)
        if nargs not in ('?', '+', '*', REMAINDER):
            limit = int(nargs)

        return [cast_value(cast, x) for x in val[:limit]]

    return wrap


class AnyBlokActionsContainer:

    def add_argument(self, *args, **kwargs):
        """Overload the method to add the entry in the configuration dict"""
        default = kwargs.pop('default', None)
        nargs = kwargs.get('nargs', None)
        type = kwargs.get('type')
        arg = super(AnyBlokActionsContainer, self).add_argument(
            *args, **kwargs)
        if type is None:
            if kwargs.get('action') == 'store_true':
                type = bool
                if default is None:
                    default = False
            else:
                type = str

        dest = arg.dest
        if nargs:
            type = nargs_type(dest, nargs, type)

        Configuration.add_argument(dest, default, type=type)
        return arg

    def set_defaults(self, **kwargs):
        """Overload the method to update the entry in the configuration dict"""
        super(AnyBlokActionsContainer, self).set_defaults(**kwargs)
        for dest, default in kwargs.items():
            if not Configuration.has(dest):
                raise KeyError('Unexisting option %s' % dest)

            Configuration.set(dest, default)

    def add_argument_group(self, *args, **kwargs):
        group = AnyBlokArgumentGroup(self, *args, **kwargs)
        self._action_groups.append(group)
        return group


class AnyBlokArgumentParser(AnyBlokActionsContainer, ArgumentParser):
    """Over load the argparse in the goal to define type directly by
    Configuration class for argparse and config files
    """


class AnyBlokArgumentGroup(AnyBlokActionsContainer, _ArgumentGroup):
    """Over load the argparse group in the goal to define type directly by
    Configuration class for argparse and config files
    """


def AnyBlokPlugin(import_definition):
    """Callable to cast import string definition to object

    ::

        path = AnyBlokPlugin('sys:path')
        //
        Registry = AnyBlokPlugin('anyblok.registry:Registry')

    :param import_definition: string of the object to import
    :rtype: imported object
    """
    if not isinstance(import_definition, str):
        return import_definition

    import_path, import_name = import_definition.split(':')
    module = __import__(import_path, fromlist=[import_name])
    if hasattr(module, import_name):
        return getattr(module, import_name)

    raise ImportError("%s does not exist in %s" % (import_name, import_path))


class ConfigOption:

    def __init__(self, value, type):
        self.type = type
        self.set(value)

    def get(self):
        return self.value

    def set(self, value):
        self.value = cast_value(self.type, value)


def getParser(**kwargs):
    """ Return a parser

    :param description: label of the configuration help
    :rtype: ``ArgumentParser`` instance
    """
    return AnyBlokArgumentParser(**kwargs)


class Configuration:
    """ ``Configuration`` is used to define the options of the real argparse
    and its default values. Each application or blok can declare needed
    options here.

    This class stores three attributes:

    * groups: lists of options indexed
    * labels: if a group has got a label then all the options in group are
      gathered in a parser group
    * configuration: result of the ``Configuration`` after loading

    """

    groups = {}
    labels = {}
    configuration = {}
    applications = {
        'default': {
            'description': "[options] -- other arguments",
            'configuration_groups': ['config', 'database'],
        },
    }

    @classmethod
    def add_application_properties(cls, application, new_groups,
                                   add_default_group=True, **kwargs):
        """Add argparse properties for an application

        If the application does not exist, the application will be added in
        the applications properties storage.

        new_groups: extend the existing groups defined. If no group is defined
        before, this extend the groups of the default application.

        :param application: the name of the application
        :param new_groups: list of the configuration groups
        :param add_default_group: if True the default groups will be add
        :params kwargs: other properties to change
        """
        if application not in cls.applications:
            cls.applications[application] = {}
        app = cls.applications[application]
        if kwargs:
            app.update(kwargs)

        if new_groups:
            if 'configuration_groups' not in app:
                app['configuration_groups'] = []
                if add_default_group:
                    app['configuration_groups'].extend(
                        cls.applications['default']['configuration_groups'])

            cg = app['configuration_groups']
            for new_group in new_groups:
                if new_group not in cg:
                    cg.append(new_group)

    @classmethod
    def add(cls, group, label=None, function_=None,
            must_be_loaded_by_unittest=False):
        """ Add a function in a group.

        The function must have two arguments:

        * ``parser``: the parser instance of argparse
        * ``default``: A dict with the default value

        This function is called to know what the options of this must do.
        You can declare this group:

        * either by calling the ``add`` method as a function::

            def foo(parser, default):
                pass

            Configuration.add('create-db', function_=foo)

        * or by calling the ``add`` method as a decorator::

            @Configuration.add('create-db')
            def bar(parser, default):
                pass

        By default the group is unnamed, if you want a named group, you must
        set the ``label`` attribute::

            @Configuration.add('create-db', label="Name of the group")
            def bar(parser, default):
                pass

        :param group: group is a set of parser option
        :param label: If the group has a label then all the functions in the
            group are put in group parser
        :param function_: function to add
        :param must_be_loaded_by_unittest: unittest call this function to init
            configuration of AnyBlok for run unittest"
        """
        if group not in cls.groups:
            cls.groups[group] = []

        if label:
            cls.labels[group] = label

        if function_:
            if function_ not in cls.groups[group]:
                function_.must_be_loaded_by_unittest = \
                    must_be_loaded_by_unittest
                cls.groups[group].append(function_)
        else:

            def wrapper(function):
                if function not in cls.groups[group]:
                    function.must_be_loaded_by_unittest = \
                        must_be_loaded_by_unittest
                    cls.groups[group].append(function)
                return function

            return wrapper

    @classmethod
    def has(cls, option):
        """ Check if the option exist in the configuration dict

        Return True if the option is in the configuration dict and the
        value is not None. A None value is diferent that a ConfigOption with
        None value

        :param opt: option key to check
        :rtype: boolean True is exist
        """
        if option in cls.configuration and cls.configuration[option]:
            return True

        return False

    @classmethod
    def get(cls, opt, default=None):
        """ Get a value from the configuration dict after loading

        After the loading of the application, all the options are saved in the
        Configuration. And all the applications have free access to
        these options::

            from anyblok._configuration import Configuration

            database = Configuration.get('db_name')

        ..warning::

            Some options are used as a default value not real value, such
            as the db_name

        :param opt: name of the option
        :param default: default value if the option doesn't exist
        """
        if cls.has(opt):
            return cls.configuration[opt].get()

        return default

    @classmethod
    def set(cls, opt, value):
        """ Set a value to the configuration dict

        :param opt: name of the option
        :param value: value to set
        """

        try:
            if opt in cls.configuration:
                cls.configuration[opt].set(value)
            else:
                cls.add_argument(opt, value, type(value))
        except Exception as e:
            logger.exception("Error durring set the value %r on the option "
                             "%r" % (value, opt))
            raise e

    @classmethod
    def update(cls, *args, **kwargs):
        if args:
            if len(args) > 1:
                raise ConfigurationException("Waiting only one dict")

            if not isinstance(args[0], dict):
                raise ConfigurationException("Waiting a dict")

            for k, v in args[0].items():
                cls.set(k, v)

        for k, v in kwargs.items():
            cls.set(k, v)

    @classmethod
    def add_argument(cls, key, value, type=str):
        cls.configuration[key] = ConfigOption(value, type)

    @classmethod
    def remove_label(cls, group):
        """ Remove an existing label

        The goal of this function is to remove an existing label of a specific
        group::

            @Configuration.add('create-db', label="Name of the group")
            def bar(parser, defaul):
                pass

            Configuration.remove_label('create-db')

        :param group: group is a set of parser option
        """
        if group in cls.labels:
            del cls.labels[group]

    @classmethod
    def remove(cls, group, function_):
        """ Remove an existing function

        If your application inherits some unwanted options from a specific
        function, you can unlink this function::

            def foo(opt, default):
                pass

            Configuration.add('create-db', function_=foo)
            Configuration.remove('create-db', function_=foo)

        :param group: group is a set of parser option
        :param function_: function to add
        """
        if group in cls.groups:
            if function_ in cls.groups[group]:
                cls.groups[group].remove(function_)

    @classmethod
    @log(logger)
    def load_config_for_test(cls):
        """Load the argparse configuration need for the unittest"""
        if not cls.configuration:
            parser = getParser()
            for group in cls.groups.keys():
                for fnct in cls.groups[group]:
                    if (
                        fnct.must_be_loaded_by_unittest or
                        group in ('plugins',)
                    ):
                        fnct(parser)

    @classmethod
    @log(logger, level='debug')
    def load(cls, application, useseparator=False, **kwargs):
        """ Load the argparse definition and parse them

        :param application: name of the application
        :param useseparator: boolean(default False)
        :param \**kwargs: ArgumentParser named arguments
        """

        sep = len(sys.argv)
        our_argv = sys.argv[1:]
        description = {}
        if application in cls.applications:
            description.update(cls.applications[application])
        else:
            description.update(cls.applications['default'])

        description.update(kwargs)
        configuration_groups = description.pop(
            'configuration_groups',
            cls.applications['default']['configuration_groups'])
        if 'plugins' not in configuration_groups:
            configuration_groups.append('plugins')

        if useseparator:
            parser = getParser(**description)

            try:
                sep = sys.argv.index('--')
                our_argv = sys.argv[1:sep]
            except ValueError:
                pass
        else:
            parser = getParser(**description)

        cls._load(parser, configuration_groups)
        args = parser.parse_args(our_argv)
        if sep is not None:
            del sys.argv[1:sep+1]

        cls.parse_options(args)

    @classmethod
    def _load(cls, parser, configuration_groups):
        """ Load the argparse definition and parse them

        :param description: description of configuration
        :param configuration_groups: list configuration groupe to load
        :param useseparator: boolean(default False)
        """
        if configuration_groups is None:
            return

        for group in cls.groups:
            if group not in configuration_groups:
                continue

            label = cls.labels.get(group)
            if label:
                g = parser.add_argument_group(label)
            else:
                g = parser

            for function in cls.groups[group]:
                function(g)

    @classmethod
    def initialize_logging(cls):
        level = {
            'NOTSET': NOTSET,
            'DEBUG': DEBUG,
            'INFO': INFO,
            'WARNING': WARNING,
            'ERROR': ERROR,
            'CRITICAL': CRITICAL
        }.get(cls.get('logging_level'))
        logging_configfile = cls.get('logging_configfile')
        json_logging_configfile = cls.get('json_logging_configfile')
        yaml_logging_configfile = cls.get('yaml_logging_configfile')
        logging_level_qualnames = cls.get('logging_level_qualnames')

        if logging_configfile:
            config.fileConfig(logging_configfile)
        elif json_logging_configfile:
            with open(json_logging_configfile, 'rt') as f:
                configfile = json.load(f.read())
                config.dictConfig(configfile)
        elif yaml_logging_configfile:
            with open(yaml_logging_configfile, 'rt') as f:
                configfile = yaml.load(f.read())
                config.dictConfig(configfile)

        if level:
            basicConfig(level=level)
            if logging_level_qualnames:
                qualnames = set(x.split('.')[0]
                                for x in Logger.manager.loggerDict.keys()
                                if x in logging_level_qualnames)
            else:
                qualnames = set(x.split('.')[0]
                                for x in Logger.manager.loggerDict.keys())

            for qualname in qualnames:
                getLogger(qualname).setLevel(level)

    @classmethod
    def parse_configfile(cls, configfile, required):
        cur_cwd = os.getcwd()
        configfile = os.path.abspath(configfile)
        print('Loading config file %r' % configfile)
        if not isfile(configfile):
            if required:
                raise ConfigurationException(
                    "No such file or not a regular file: %r " % configfile)

            return

        cwd_file, file_name = os.path.split(configfile)
        configuration = {}
        try:
            os.chdir(cwd_file)
            cparser = ConfigParser()
            cparser.read(configfile)
            sections = cparser.items('AnyBlok')

            for opt, value in sections:
                if opt in ('logging_configfile', 'json_logging_configfile',
                           'yaml_logging_configfile'):
                    if value:
                        value = os.path.abspath(value)

                configuration[opt] = value

            if 'extend' in configuration:
                extend = configuration.pop('extend')
                if extend:
                    cls.parse_configfile(extend, True)

        finally:
            os.chdir(cur_cwd)

        cls.update(**configuration)

    @classmethod
    def parse_options(cls, arguments):
        if arguments._get_args():
            raise ConfigurationException(
                'Positional arguments are forbidden')

        ad = AppDirs('AnyBlok')
        # load the global configuration file
        cls.parse_configfile(
            join(ad.site_config_dir, 'conf.cfg'), False)
        # load the user configuration file
        cls.parse_configfile(
            join(ad.user_config_dir, 'conf.cfg'), False)
        if 'configfile' in dict(arguments._get_kwargs()).keys():
            if arguments.configfile:
                cls.parse_configfile(arguments.configfile, True)

        for opt, value in arguments._get_kwargs():
            if opt not in cls.configuration or value:
                cls.set(opt, value)

        if 'logging_level' in cls.configuration:
            cls.initialize_logging()


@Configuration.add('plugins', label='Plugins',
                   must_be_loaded_by_unittest=True)
def add_plugins(group):
    group.add_argument('--registry-cls', dest='Registry', type=AnyBlokPlugin,
                       default='anyblok.registry:Registry',
                       help="Registry class to use")
    group.add_argument('--migration-cls', dest='Migration',
                       type=AnyBlokPlugin,
                       default='anyblok.migration:Migration',
                       help="Migration class to use")
    group.add_argument('--get-url-fnct', dest='get_url',
                       type=AnyBlokPlugin,
                       default='anyblok.config:get_url',
                       help="get_url function to use")


@Configuration.add('config')
def add_configuration_file(parser):
    parser.add_argument('-c', dest='configfile',
                        help="Relative path of the config file")
    parser.add_argument('--without-auto-migration', dest='withoutautomigration',
                        action='store_true')
    parser.add_argument('--isolation-level',
                        default="READ_COMMITTED",
                        choices=["SERIALIZABLE", "REPEATABLE_READ",
                                 "READ_COMMITTED", "READ_UNCOMMITTED",
                                 "AUTOCOMMIT"])
    parser.add_argument('--default-timezone',
                        default=os.environ.get('ANYBLOK_DEFAULT_TIMEZONE'),
                        help="default timezone use by naive datetime "
                             "(by default use the timezone of the serveur")


@Configuration.add('database', label="Database",
                   must_be_loaded_by_unittest=True)
def add_database(group):
    group.add_argument('--db-name',
                       default=os.environ.get('ANYBLOK_DATABASE_NAME'),
                       help="Name of the database")
    group.add_argument('--db-url',
                       default=os.environ.get('ANYBLOK_DATABASE_URL'),
                       help="Complete URL for connection with the database")
    group.add_argument('--db-driver-name',
                       default=os.environ.get('ANYBLOK_DATABASE_DRIVER'),
                       help="the name of the database backend. This name "
                            "will correspond to a module in "
                            "sqlalchemy/databases or a third party plug-in")
    group.add_argument('--db-user-name',
                       default=os.environ.get('ANYBLOK_DATABASE_USER'),
                       help="The user name")
    group.add_argument('--db-password',
                       default=os.environ.get('ANYBLOK_DATABASE_PASSWORD'),
                       help="database password")
    group.add_argument('--db-host',
                       default=os.environ.get('ANYBLOK_DATABASE_HOST'),
                       help="The name of the host")
    group.add_argument('--db-port', type=int,
                       default=os.environ.get('ANYBLOK_DATABASE_PORT'),
                       help="The port number")
    group.add_argument('--db-echo', action="store_true",
                       default=(os.environ.get(
                           'ANYBLOK_DATABASE_ECHO') or False))
    group.add_argument('--db-echo-pool', action="store_true", default=False)
    group.add_argument('--db-max-overflow', type=int, default=10)
    group.add_argument('--db-pool-size', type=int, default=5)
    group.add_argument('--default-encrypt-key',
                       default=os.environ.get('ANYBLOK_ENCRYPT_KEY'),
                       help=("Default ey definition to encrypt column with "
                             "encryp_key=True"))


@Configuration.add('create_db', must_be_loaded_by_unittest=True)
def add_create_database(group):
    group.add_argument(
        '--db-template-name',
        default=os.environ.get('ANYBLOK_DATABASE_TEMPLATE_NAME'),
        help="Name of the template")


@Configuration.add('install-bloks')
def add_install_bloks(parser):
    parser.add_argument('--install-bloks', nargs="+", help="blok to install")
    parser.add_argument('--install-all-bloks', action='store_true')
    parser.add_argument('--test-blok-at-install', action='store_true')


@Configuration.add('uninstall-bloks')
def add_uninstall_bloks(parser):
    parser.add_argument('--uninstall-bloks', nargs="+",
                        help="bloks to uninstall")


@Configuration.add('install-or-update-bloks')
def add_install_or_update_bloks(parser):
    parser.add_argument('--install-or-update-bloks', nargs="+",
                        help="bloks to install or update")


@Configuration.add('update-bloks', label='Update database')
def add_update_bloks(parser):
    parser.add_argument('--update-bloks', nargs="+", help="bloks to update")
    parser.add_argument('--update-all-bloks', action='store_true')
    parser.add_argument('--reinit-all', action='store_true')
    parser.add_argument('--reinit-tables', action='store_true')
    parser.add_argument('--reinit-columns', action='store_true')
    parser.add_argument('--reinit-indexes', action='store_true')
    parser.add_argument('--reinit-constraints', action='store_true')


@Configuration.add('interpreter')
def add_interpreter(parser):
    parser.add_argument('--script', dest='python_script',
                        help="Python script to execute")


@Configuration.add('schema', label="Schema options")
def add_schema(group):
    try:
        from graphviz.files import FORMATS
    except ImportError:
        from graphviz.backend import FORMATS

    group.add_argument('--schema-format',
                       default='png', choices=tuple(FORMATS))


@Configuration.add('doc', label="Doc options")
def add_doc(group):
    group.add_argument('--doc-format',
                       default='RST', choices=('RST', 'UML', 'SQL'))
    group.add_argument('--doc-output',
                       default='anyblok-documentation')
    group.add_argument('--doc-wanted-models', nargs='+',
                       help='Detail only these models')
    group.add_argument('--doc-unwanted-models', nargs='+',
                       help='No detail these models')


@Configuration.add('unittest', label="Unittest")
def add_unittest(group):
    group.add_argument('--selected-bloks', nargs='+',
                       help="Name of the bloks to test")
    group.add_argument('--unwanted-bloks', nargs='+',
                       help="Name of the bloks to no test")


@Configuration.add('logging', label="Logging")
def add_logging(group):
    group.add_argument('--logging-level', dest='logging_level',
                       choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING',
                                'ERROR', 'CRITICAL'])
    group.add_argument('--logging-level-qualnames', nargs='+',
                       help="Limit the log level on a qualnames list")
    group.add_argument('--logging-config-file', dest='logging_configfile',
                       help="Relative path of the logging config file")
    group.add_argument('--logging-json-config-file',
                       dest='json_logging_configfile',
                       help="Relative path of the logging config file (json). "
                            "Only if the logging config file doesn't filled")
    group.add_argument('--logging-yaml-config-file',
                       dest='yaml_logging_configfile',
                       help="Relative path of the logging config file (yaml). "
                            "Only if the logging and json config file doesn't "
                            "filled")


@Configuration.add('preload', label="Preload")
def define_preload_option(group):
    group.add_argument('--databases', dest='db_names', nargs="+",
                       help='List of the database allow to be load')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase
from sqlalchemy import Integer as SA_Integer, String as SA_String
from sqlalchemy.exc import StatementError
from anyblok import Declarations
from anyblok.field import FieldException
from anyblok.column import (
    Column, Boolean, Json, String, BigInteger, Text, Selection, Date, DateTime,
    Time, Interval, Decimal, Float, LargeBinary, Integer, Sequence, Color,
    Password, UUID, URL, PhoneNumber, Email, Country)
from unittest import skipIf

try:
    import cryptography  # noqa
    has_cryptography = True
except Exception:
    has_cryptography = False

try:
    import passlib  # noqa
    has_passlib = True
except Exception:
    has_passlib = False

try:
    import colour  # noqa
    has_colour = True
except Exception:
    has_colour = False

try:
    import furl  # noqa
    has_furl = True
except Exception:
    has_furl = False


try:
    import phonenumbers  # noqa
    has_phonenumbers = True
except Exception:
    has_phonenumbers = False

try:
    import pycountry  # noqa
    has_pycountry = True
    from sqlalchemy_utils import PhoneNumber as PN
except Exception:
    has_pycountry = False


Model = Declarations.Model
register = Declarations.register


class OneColumn(Column):
    sqlalchemy_type = SA_Integer


class TestColumn(TestCase):

    def test_forbid_instance(self):
        try:
            Column()
            self.fail("Column mustn't be instanciated")
        except FieldException:
            pass

    def test_without_label(self):
        column = OneColumn()
        column.get_sqlalchemy_mapping(None, None, 'a_column', None)
        self.assertEqual(column.label, 'A column')


def simple_column(ColumnType=None, **kwargs):

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        col = ColumnType(**kwargs)


def column_with_foreign_key():

    @register(Model)
    class Test:

        name = String(primary_key=True)

    @register(Model)
    class Test2:

        id = Integer(primary_key=True)
        test = String(foreign_key=Model.Test.use('name'))


class TestColumns(DBTestCase):

    def test_column_with_type_in_kwargs(self):
        self.init_registry(
            simple_column, ColumnType=Integer, type_=Integer)

    def test_column_with_db_column_name_in_kwargs(self):
        registry = self.init_registry(simple_column, ColumnType=Integer,
                                      db_column_name='another_name')
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)
        res = registry.execute('select id from test where another_name=1')
        self.assertEqual(res.fetchone()[0], test.id)

    def test_column_with_foreign_key(self):
        registry = self.init_registry(column_with_foreign_key)
        registry.Test.insert(name='test')
        registry.Test2.insert(test='test')

    def test_integer(self):
        registry = self.init_registry(simple_column, ColumnType=Integer)
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)

    def test_integer_str_foreign_key(self):
        registry = self.init_registry(
            simple_column, ColumnType=Integer, foreign_key='Model.Test=>id')
        test = registry.Test.insert()
        test2 = registry.Test.insert(col=test.id)
        self.assertEqual(test2.col, test.id)

    def test_big_integer(self):
        registry = self.init_registry(simple_column, ColumnType=BigInteger)
        test = registry.Test.insert(col=1)
        self.assertEqual(test.col, 1)

    def test_Float(self):
        registry = self.init_registry(simple_column, ColumnType=Float)
        test = registry.Test.insert(col=1.0)
        self.assertEqual(test.col, 1.0)

    def test_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert(col=D('1.0'))
        self.assertEqual(test.col, D('1.0'))

    def test_setter_decimal(self):
        from decimal import Decimal as D

        registry = self.init_registry(simple_column, ColumnType=Decimal)
        test = registry.Test.insert()
        test.col = '1.0'
        self.assertEqual(test.col, D('1.0'))

    def test_boolean(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean)
        test = registry.Test.insert(col=True)
        self.assertEqual(test.col, True)

    def test_boolean_with_default(self):
        registry = self.init_registry(simple_column, ColumnType=Boolean,
                                      default=False)
        test = registry.Test.insert()
        self.assertEqual(test.col, False)

    def test_string(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_string_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_string_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_string_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=String)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    @skipIf(not has_cryptography, "cryptography is not installed")
    def test_string_with_encrypt_key(self):
        registry = self.init_registry(simple_column, ColumnType=String,
                                      encrypt_key='secretkey')
        test = registry.Test.insert(col='col')
        registry.session.commit()
        self.assertEqual(test.col, 'col')
        res = registry.execute('select col from test where id = %s' % test.id)
        res = res.fetchall()[0][0]
        self.assertNotEqual(res, 'col')

    @skipIf(not has_cryptography, "cryptography is not installed")
    def test_datetime_with_encrypt_key(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime,
                                      encrypt_key='secretkey')
        test = registry.Test.insert(col=now)
        registry.session.commit()
        self.assertEqual(test.col, now)

    def test_string_with_size(self):
        registry = self.init_registry(
            simple_column, ColumnType=String, size=100)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    @skipIf(not has_passlib, "passlib is not installed")
    def test_password(self):
        registry = self.init_registry(simple_column, ColumnType=Password,
                                      crypt_context={'schemes': ['md5_crypt']})
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')
        self.assertNotEqual(
            registry.execute('Select col from test').fetchone()[0], 'col')

    def test_text(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col='col')
        self.assertEqual(test.col, 'col')

    def test_text_with_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert(col=False)
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_text_set_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        test.col = False
        self.assertEqual(test.col, False)
        self.registry.flush()
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_text_query_False(self):
        registry = self.init_registry(simple_column, ColumnType=Text)
        test = registry.Test.insert()
        self.registry.Test.query().filter_by(id=test.id).update({'col': False})
        self.assertEqual(test.col, False)
        self.registry.expire(test, ['col'])
        self.assertEqual(test.col, '')

    def test_date(self):
        from datetime import date

        now = date.today()
        registry = self.init_registry(simple_column, ColumnType=Date)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_datetime(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_datetime_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=None)
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S.%f'))
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now.strftime('%Y-%m-%d %H:%M:%S'))
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now
        self.assertEqual(test.col, now)

    def test_datetime_by_property_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = None
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%z')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S.%f')
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4_by_property(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        test.col = now.strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=now))
        self.assertEqual(test.col, now)

    def test_datetime_by_query_none_value(self):
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(dict(col=None))
        self.assertIsNone(test.col)

    def test_datetime_str_conversion_1_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%z')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_2_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_3_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S.%f')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now)

    def test_datetime_str_conversion_4_by_query(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert()
        registry.Test.query().update(
            dict(col=now.strftime('%Y-%m-%d %H:%M:%S')))
        registry.expire(test, ['col'])
        self.assertEqual(test.col, now.replace(microsecond=0))

    def test_datetime_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(Test.query().filter(Test.col == now).one(), test)

    def test_datetime_str_conversion_1_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = datetime.datetime.now().replace(tzinfo=timezone)
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%z')).one(),
            test)

    def test_datetime_str_conversion_2_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f%Z')).one(),
            test)

    def test_datetime_str_conversion_3_by_query_filter(self):
        import datetime
        import time
        import pytz

        timezone = pytz.timezone(time.tzname[0])
        now = timezone.localize(datetime.datetime.now())
        registry = self.init_registry(simple_column, ColumnType=DateTime)
        test = registry.Test.insert(col=now)
        Test = registry.Test
        self.assertIs(
            Test.query().filter(
                Test.col == now.strftime('%Y-%m-%d %H:%M:%S.%f')).one(), test)

    def test_datetime_without_auto_update_1(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime()
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        self.assertIsNone(test.update_at)
        test.val = 'other'
        registry.flush()
        self.assertIsNone(test.update_at)

    def test_datetime_without_auto_update_2(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime(auto_update=False)
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        self.assertIsNone(test.update_at)
        test.val = 'other'
        registry.flush()
        self.assertIsNone(test.update_at)

    def test_datetime_with_auto_update(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                id = Integer(primary_key=True)
                update_at = DateTime(auto_update=True)
                val = String()

        registry = self.init_registry(add_in_registry)
        test = registry.Test.insert(val='first add')
        self.assertIsNone(test.update_at)
        test.val = 'other'
        registry.flush()
        self.assertIsNotNone(test.update_at)

    def test_datetime_with_default_timezone_tz(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        registry = self.init_registry(
            simple_column, ColumnType=DateTime,
            default_timezone=timezone)
        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        self.assertIs(field.default_timezone, timezone)

        test = registry.Test.insert(col=now)
        self.assertIs(test.col.tzinfo.zone, timezone.zone)

    def test_datetime_with_default_timezone_str(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        registry = self.init_registry(
            simple_column, ColumnType=DateTime,
            default_timezone='Asia/Tokyo')
        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        self.assertIs(field.default_timezone, timezone)

        test = registry.Test.insert(col=now)
        self.assertIs(test.col.tzinfo.zone, timezone.zone)

    def test_datetime_with_default_global_timezone_str(self):
        import datetime
        import pytz

        timezone = pytz.timezone('Asia/Tokyo')
        now = datetime.datetime.now()
        with DBTestCase.Configuration(default_timezone='Asia/Tokyo'):
            registry = self.init_registry(simple_column, ColumnType=DateTime)

        field = registry.loaded_namespaces_first_step['Model.Test']['col']
        self.assertIs(field.default_timezone, timezone)

        test = registry.Test.insert(col=now)
        self.assertIs(test.col.tzinfo.zone, timezone.zone)

    def test_interval(self):
        from datetime import timedelta

        dt = timedelta(days=5)
        registry = self.init_registry(simple_column, ColumnType=Interval)
        test = registry.Test.insert(col=dt)
        self.assertEqual(test.col, dt)

    def test_time(self):
        from datetime import time

        now = time()
        registry = self.init_registry(simple_column, ColumnType=Time)
        test = registry.Test.insert(col=now)
        self.assertEqual(test.col, now)

    def test_large_binary(self):
        from os import urandom

        blob = urandom(100000)
        registry = self.init_registry(simple_column, ColumnType=LargeBinary)

        test = registry.Test.insert(col=blob)
        self.assertEqual(test.col, blob)

    def test_selection(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        test = registry.Test.insert(col=SELECTIONS[0][0])
        self.assertEqual(test.col, SELECTIONS[0][0])
        self.assertEqual(test.col.label, SELECTIONS[0][1])
        test = registry.Test.query().first()
        self.assertEqual(test.col, SELECTIONS[0][0])
        self.assertEqual(test.col.label, SELECTIONS[0][1])
        with self.assertRaises(FieldException):
            test.col = 'bad value'

    def test_selection_with_none_value(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        t = registry.Test.insert()
        self.assertIsNone(t.col)

    def test_selection_change_by_query(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        with self.assertRaises(StatementError):
            registry.Test.query().update({'col': 'bad value'})

    def test_selection_like_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        Test = registry.Test
        t = Test.insert(col=SELECTIONS[0][0])
        t1 = Test.query().filter(Test.col.like('%admin%')).one()
        self.assertIs(t, t1)

    def test_selection_key_other_than_str(self):
        SELECTIONS = [
            (0, 'Admin'),
            (1, 'Regular user')
        ]

        with self.assertRaises(FieldException):
            self.init_registry(
                simple_column, ColumnType=Selection, selections=SELECTIONS)

    def test_selection_comparator(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        registry = self.init_registry(
            simple_column, ColumnType=Selection, selections=SELECTIONS)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    def test_selection_use_method(self):
        SELECTIONS = [
            ('admin', 'Admin'),
            ('regular-user', 'Regular user')
        ]

        def add_selection():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                col = Selection(selections='get_selection')

                @classmethod
                def get_selection(cls):
                    return SELECTIONS

        registry = self.init_registry(add_selection)
        registry.Test.insert(col=SELECTIONS[0][0])
        registry.Test.query().filter(
            registry.Test.col.in_(['admin', 'regular-user'])).first()

    def test_json(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        val = {'a': 'Test'}
        test = registry.Test.insert(col=val)
        self.assertEqual(test.col, val)

    def test_json_update(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        test = registry.Test.insert(col={'a': 'test'})
        test.col['b'] = 'test'
        self.assertEqual(test.col, {'a': 'test', 'b': 'test'})

    def test_json_simple_filter(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col={'a': 'test'})
        Test.insert(col={'a': 'test'})
        Test.insert(col={'b': 'test'})
        self.assertEqual(
            Test.query().filter(
                Test.col['a'].cast(SA_String) == '"test"').count(),
            2)

    def test_json_null(self):
        registry = self.init_registry(simple_column, ColumnType=Json)
        Test = registry.Test
        Test.insert(col=None)
        Test.insert(col=None)
        Test.insert(col={'a': 'test'})
        self.assertEqual(Test.query().filter(Test.col.is_(None)).count(), 2)
        self.assertEqual(Test.query().filter(Test.col.isnot(None)).count(), 1)

    def test_add_default_classmethod(self):
        val = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                @classmethod
                def get_val(cls):
                    return val

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, val)

    def test_add_default_without_classmethod(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default="get_val")

                def get_val(cls):
                    return value

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, "get_val")

    def test_add_default_by_var(self):
        value = 'test'

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default=value)

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, value)

    def test_add_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='value')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, 'value')

    def test_add_field_as_default(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                val = String(default='val')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        self.assertEqual(t.val, 'val')

    def test_sequence(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence)
        self.assertEqual(registry.Test.insert().col, "1")
        self.assertEqual(registry.Test.insert().col, "2")
        self.assertEqual(registry.Test.insert().col, "3")
        self.assertEqual(registry.Test.insert().col, "4")
        Seq = registry.System.Sequence
        self.assertEqual(
            Seq.query().filter(Seq.code == 'Model.Test=>col').count(), 1)

    def test_sequence_with_primary_key(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence,
                                      primary_key=True)
        self.assertEqual(registry.Test.insert().col, "1")
        self.assertEqual(registry.Test.insert().col, "2")

    def test_sequence_with_code_and_formater(self):
        registry = self.init_registry(simple_column, ColumnType=Sequence,
                                      code="SO", formater="{code}-{seq:06d}")
        self.assertEqual(registry.Test.insert().col, "SO-000001")
        self.assertEqual(registry.Test.insert().col, "SO-000002")
        self.assertEqual(registry.Test.insert().col, "SO-000003")
        self.assertEqual(registry.Test.insert().col, "SO-000004")
        Seq = registry.System.Sequence
        self.assertEqual(Seq.query().filter(Seq.code == 'SO').count(), 1)

    def test_sequence_with_foreign_key(self):
        with self.assertRaises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               foreign_key=Model.System.Model.use('name'))

    def test_sequence_with_default(self):
        with self.assertRaises(FieldException):
            self.init_registry(simple_column, ColumnType=Sequence,
                               default='default value')

    @skipIf(not has_colour, "colour is not installed")
    def test_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert(col=color)
        self.assertEqual(test.col.hex, colour.Color(color).hex)

    @skipIf(not has_colour, "colour is not installed")
    def test_setter_color(self):
        color = '#F5F5F5'
        registry = self.init_registry(simple_column, ColumnType=Color)
        test = registry.Test.insert()
        test.col = color
        self.assertEqual(test.col.hex, colour.Color(color).hex)

    def test_uuid_binary_1(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_3(self):
        from uuid import uuid3, NAMESPACE_DNS
        uuid = uuid3(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_4(self):
        from uuid import uuid4
        uuid = uuid4()
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_binary_5(self):
        from uuid import uuid5, NAMESPACE_DNS
        uuid = uuid5(NAMESPACE_DNS, 'python.org')
        registry = self.init_registry(simple_column, ColumnType=UUID)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    def test_uuid_char32(self):
        from uuid import uuid1
        uuid = uuid1()
        registry = self.init_registry(simple_column, ColumnType=UUID,
                                      binary=False)
        test = registry.Test.insert(col=uuid)
        self.assertIs(test.col, uuid)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        self.assertEqual(test.col, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_setter_URL(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert()
        test.col = f
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_2(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        test = registry.Test.insert(col=f)
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_3(self):
        f = 'http://doc.anyblok.org'
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        self.assertEqual(test.col.url, f)

    @skipIf(not has_furl, "furl is not installed")
    def test_URL_4(self):
        f = furl.furl('http://doc.anyblok.org')
        registry = self.init_registry(simple_column, ColumnType=URL)
        registry.Test.insert(col=f)
        Test = registry.Test
        test = Test.query().filter(Test.col == f).one()
        self.assertEqual(test.col.url, f.url)

    @skipIf(not has_phonenumbers, "phonenumbers is not installed")
    def test_phonenumbers_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert(col='+120012301')
        self.assertEqual(test.col.national, '20012301')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0],
            '+120012301')

    @skipIf(not has_phonenumbers, "phonenumbers is not installed")
    def test_phonenumbers_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert()
        test.col = '+120012301'
        self.assertEqual(test.col.national, '20012301')
        registry.flush()
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0],
            '+120012301')

    @skipIf(not has_phonenumbers, "phonenumbers is not installed")
    def test_phonenumbers_obj_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        col = PN("+120012301", None)
        test = registry.Test.insert(col=col)
        self.assertEqual(test.col, col)
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0],
            '+120012301')

    @skipIf(not has_phonenumbers, "phonenumbers is not installed")
    def test_phonenumbers_obj_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=PhoneNumber)
        test = registry.Test.insert()
        test.col = PN("+120012301", None)
        self.assertEqual(test.col.national, '20012301')
        registry.flush()
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0],
            '+120012301')

    def test_email_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=Email)
        test = registry.Test.insert(col='John.Smith@foo.com')
        self.assertEqual(test.col, 'john.smith@foo.com')
        self.assertEqual(
            registry.Test.query().filter_by(col='John.Smith@foo.com').count(),
            1)

    def test_email_at_setter(self):
        registry = self.init_registry(simple_column, ColumnType=Email)
        test = registry.Test.insert()
        test.col = 'John.Smith@foo.com'
        self.assertEqual(test.col, 'john.smith@foo.com')
        self.assertEqual(
            registry.Test.query().filter_by(col='John.Smith@foo.com').count(),
            1)

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_insert(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        test = registry.Test.insert(col='FR')
        self.assertIs(test.col, pycountry.countries.get(alpha_2='FR'))
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_insert_with_alpha_3(self):
        registry = self.init_registry(
            simple_column, ColumnType=Country, mode='alpha_3')
        test = registry.Test.insert(col='FRA')
        self.assertIs(test.col, pycountry.countries.get(alpha_2='FR'))
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_insert_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        fr = pycountry.countries.get(alpha_2='FR')
        test = registry.Test.insert(col=fr)
        self.assertIs(test.col, fr)
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_update(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        test = registry.Test.insert()
        test.col = 'FR'
        registry.flush()
        self.assertIs(test.col, pycountry.countries.get(alpha_2='FR'))
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_update_with_alpha_3(self):
        registry = self.init_registry(
            simple_column, ColumnType=Country, mode='alpha_3')
        test = registry.Test.insert()
        test.col = 'FRA'
        registry.flush()
        self.assertIs(test.col, pycountry.countries.get(alpha_2='FR'))
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_at_update_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        fr = pycountry.countries.get(alpha_2='FR')
        test = registry.Test.insert()
        test.col = fr
        registry.flush()
        self.assertIs(test.col, fr)
        self.assertEqual(test.col.name, 'France')
        self.assertEqual(
            registry.execute('Select col from test').fetchone()[0], 'FRA')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_query_is_with_object(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        Test = registry.Test
        fr = pycountry.countries.get(alpha_2='FR')
        Test.insert(col='FR')
        self.assertEqual(Test.query().filter(Test.col == fr).count(), 1)

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_query_is_with_alpha_3(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        Test = registry.Test
        Test.insert(col='FR')
        self.assertEqual(Test.query().filter(Test.col == 'FRA').count(), 1)

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_query_insert_wrong(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        with self.assertRaises(LookupError):
            registry.Test.insert(col='WG')

    @skipIf(not has_pycountry, "pycountry is not installed")
    def test_pycoundtry_query_insert_by_wrong_query(self):
        registry = self.init_registry(simple_column, ColumnType=Country)
        with self.assertRaises(Exception):
            registry.execute("insert into test (col) values ('WG2')")



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import CSVImporterException
from os import urandom
from csv import DictReader
from io import StringIO


class TestImportCSV(BlokTestCase):

    def create_importer(self, file_to_import=None, **kwargs):
        CSV = self.registry.IO.Importer.CSV
        if 'model' not in kwargs:
            kwargs['model'] = 'Model.IO.Importer'

        if file_to_import is None:
            file_to_import = urandom(100000)
        return CSV.insert(file_to_import=file_to_import, **kwargs)

    def create_csv_importer(self, **kwargs):
        CSV = self.registry.IO.Importer.CSV
        return CSV(self.create_importer(**kwargs))

    def get_file_to_import(self):
        return '''"A","B","C"\n"1","2","3"\n"4","5","6"'''.encode('utf-8')

    def test_commit_if_error_found(self):
        importer = self.create_csv_importer()
        importer.error_found.append('Mock error')
        self.assertEqual(importer.commit(), False)

    def test_commit(self):
        importer = self.create_csv_importer()
        csvfile = StringIO()
        importer.reader = DictReader(csvfile)
        self.assertEqual(importer.commit(), True)

    def test_get_reader(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        self.assertEqual(isinstance(importer.reader, DictReader), True)

    def assertNbLines(self, importer, nb_line_wanted):
        rows = [row for row in importer.reader]
        self.assertEqual(len(rows), nb_line_wanted)

    def test_consume_without_offset(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        importer.consume_offset()
        self.assertNbLines(importer, 2)

    def test_consume_offset(self):
        importer = self.create_csv_importer(
            offset=1,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        importer.consume_offset()
        self.assertNbLines(importer, 1)

    def test_consume_offset_equal(self):
        importer = self.create_csv_importer(
            offset=2,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        importer.consume_offset()
        self.assertNbLines(importer, 0)

    def test_consume_offset_greater_than_nb_lines(self):
        importer = self.create_csv_importer(
            offset=3,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        importer.consume_offset()
        self.assertNbLines(importer, 0)

    def test_consume_nb_grouped_lines(self):
        importer = self.create_csv_importer(
            nb_grouped_lines=1,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        rows = importer.consume_nb_grouped_lines()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0], {'A': '1', 'B': '2', 'C': '3'})
        rows = importer.consume_nb_grouped_lines()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0], {'A': '4', 'B': '5', 'C': '6'})

    def test_consume_nb_grouped_lines_same_number(self):
        importer = self.create_csv_importer(
            nb_grouped_lines=2,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        rows = importer.consume_nb_grouped_lines()
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0], {'A': '1', 'B': '2', 'C': '3'})
        self.assertEqual(rows[1], {'A': '4', 'B': '5', 'C': '6'})

    def test_consume_nb_group_lines_greater_than_nb_lines(self):
        importer = self.create_csv_importer(
            nb_grouped_lines=3,
            file_to_import=self.get_file_to_import())
        importer.get_reader()
        rows = importer.consume_nb_grouped_lines()
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0], {'A': '1', 'B': '2', 'C': '3'})
        self.assertEqual(rows[1], {'A': '4', 'B': '5', 'C': '6'})

    def get_exporter_file_to_import(self, withmapping=False):
        if not withmapping:
            res = '''"model","mode"\n'''
            res += '''"Model.IO.Exporter","Model.IO.Exporter.CSV"'''
        else:
            res = '''"model/EXTERNAL_ID","mode"\n'''
            res += '''"exporter_mapping","Model.IO.Exporter.CSV"'''
            exporter = self.registry.IO.Exporter.CSV.insert(
                model="Model.IO.Importer")
            self.registry.IO.Mapping.set('exporter_mapping', exporter)

        return res.encode('utf-8')

    def get_model_file_to_import(self, withmapping=False, short=False):
        if not withmapping:
            res = '''"name","table"\n'''
            res += '''"Model.IO.Test","io_test"'''
        else:
            if short:
                res = '''"/EXTERNAL_ID","table"\n'''
            else:
                res = '''"name/EXTERNAL_ID","table"\n'''
            res += '''"exporter_mapping","io_test"'''

            model = self.registry.System.Model.insert(
                name="Model.IO.Test", table='io_other_table')
            self.registry.IO.Mapping.set('exporter_mapping', model)

        return res.encode('utf-8')

    def get_column_file_to_import(self, withmapping=False):
        if not withmapping:
            res = '''"model","name","nullable"\n'''
            res += '''"Model.IO.Importer","test","1"'''
        else:
            res = '''"model/EXTERNAL_ID","name/EXTERNAL_ID","nullable"\n'''
            res += '''"exporter_mapping","exporter_mapping","1"'''
            column = self.registry.System.Column.insert(
                model="Model.IO.Importer", name="test", nullable=False)
            self.registry.IO.Mapping.set('exporter_mapping', column)

        return res.encode('utf-8')

    def test_get_header_without_mapping_and_without_pks(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_exporter_file_to_import())
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, [])
        self.assertEqual(importer.header_external_id, None)
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 2)
        for header in ('mode', 'model'):
            self.assertIn(header, importer.fields_description.keys())
            self.assertIn(header, importer.header_fields)

    def test_get_header_with_pks(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_model_file_to_import(),
            model='Model.System.Model')
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, ['name'])
        self.assertEqual(importer.header_external_id, None)
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 2)
        for header in ('name', 'table'):
            self.assertIn(header, importer.fields_description.keys())

        self.assertNotIn('name', importer.header_fields)
        self.assertIn('table', importer.header_fields)

    def test_get_header_with_multi_pks(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_column_file_to_import(),
            model='Model.System.Column')
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, ['model', 'name'])
        self.assertEqual(importer.header_external_id, None)
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 3)
        for header in ('model', 'name', 'nullable'):
            self.assertIn(header, importer.fields_description.keys())

        self.assertNotIn('name', importer.header_fields)
        self.assertIn('nullable', importer.header_fields)

    def test_get_header_with_pks_mapping(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_model_file_to_import(withmapping=True),
            model='Model.System.Model')
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, [])
        self.assertEqual(importer.header_external_id, 'name/EXTERNAL_ID')
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 2)
        for header in ('name', 'table'):
            self.assertIn(header, importer.fields_description.keys())

        self.assertNotIn('name', importer.header_fields)
        self.assertIn('table', importer.header_fields)

    def test_get_header_with_pks_mapping_short_external_id(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_model_file_to_import(withmapping=True,
                                                         short=True),
            model='Model.System.Model')
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, [])
        self.assertEqual(importer.header_external_id, '/EXTERNAL_ID')
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 1)
        self.assertIn('table', importer.fields_description.keys())
        self.assertNotIn('name', importer.header_fields)
        self.assertIn('table', importer.header_fields)

    def test_get_header_with_multi_pks_mapping(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_column_file_to_import(withmapping=True),
            model='Model.System.Column')
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, [])
        self.assertEqual(importer.header_external_id, 'name/EXTERNAL_ID')
        self.assertEqual(importer.header_external_ids, {})
        self.assertEqual(len(importer.fields_description.keys()), 3)
        for header in ('model', 'name', 'nullable'):
            self.assertIn(header, importer.fields_description.keys())

        self.assertNotIn('name', importer.header_fields)
        self.assertIn('nullable', importer.header_fields)

    def test_get_header_with_mapping(self):
        importer = self.create_csv_importer(
            file_to_import=self.get_exporter_file_to_import(withmapping=True))
        importer.get_reader()
        importer.get_header()
        self.assertEqual(importer.header_pks, [])
        self.assertEqual(importer.header_external_id, None)
        self.assertEqual(importer.header_external_ids,
                         {'model/EXTERNAL_ID': 'model'})
        self.assertEqual(len(importer.fields_description.keys()), 2)
        for header in ('model', 'mode'):
            self.assertIn(header, importer.fields_description.keys())

        self.assertNotIn('model', importer.header_fields)
        self.assertIn('mode', importer.header_fields)

    def test_parse_row_on_error_raise(self):
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Exporter',
                                            csv_on_error='raise_now')
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Importer.fields_description(
            fields=importer.header_fields)
        with self.assertRaises(CSVImporterException):
            importer.parse_row({'mode': 'Model.IO.Exporter.CSV'})

    def test_parse_row(self):
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Exporter')
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Importer.fields_description(
            fields=importer.header_fields)
        importer.parse_row({'model': 'Model.IO.Importer',
                            'mode': 'Model.IO.Exporter.CSV'})
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_parse_row_with_raise(self):
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Exporter',
                                            csv_if_does_not_exist='raise')
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Importer.fields_description(
            fields=importer.header_fields)
        importer.parse_row({'model': 'Model.IO.Importer',
                            'mode': 'Model.IO.Exporter.CSV'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 1)

    def test_parse_row_with_pass(self):
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Exporter',
                                            csv_if_does_not_exist='pass')
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Importer.fields_description(
            fields=importer.header_fields)
        importer.parse_row({'model': 'Model.IO.Importer',
                            'mode': 'Model.IO.Exporter.CSV'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_parse_row_with_existing_pks_update(self):
        Model = self.registry.System.Model
        importer = self.create_csv_importer(model='Model.System.Model')
        importer.header_pks = ['name']
        importer.header_fields = ['table']
        importer.fields_description = Model.fields_description(
            fields=['name', 'table'])
        Model.insert(name='Model.IO.Test', table='io_test')
        importer.parse_row({'name': 'Model.IO.Test',
                            'table': 'io_test_other_table'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 1)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(importer.updated_entries[0].table,
                         'io_test_other_table')

    def test_parse_row_with_existing_pks_create(self):
        Exporter = self.registry.IO.Exporter
        importer = self.create_csv_importer(model='Model.IO.Exporter',
                                            csv_if_exist='create')
        importer.header_pks = ['id']
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Exporter.fields_description(
            fields=['id', 'model', 'mode'])
        exporter = Exporter.CSV.insert(model='Model.IO.Importer')
        importer.parse_row({'id': str(exporter.id),
                            'model': 'Model.System.Model',
                            'mode': "Model.IO.Exporter.CSV"})
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_parse_row_with_existing_pks_pass(self):
        Model = self.registry.System.Model
        importer = self.create_csv_importer(model='Model.System.Model',
                                            csv_if_exist="pass")
        importer.header_pks = ['name']
        importer.header_fields = ['table']
        importer.fields_description = Model.fields_description(
            fields=['name', 'table'])
        Model.insert(name='Model.IO.Test', table='io_test')
        importer.parse_row({'name': 'Model.IO.Test',
                            'table': 'io_test_other_table'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_parse_row_with_existing_pks_raise(self):
        Model = self.registry.System.Model
        importer = self.create_csv_importer(model='Model.System.Model',
                                            csv_if_exist="raise")
        importer.header_pks = ['name']
        importer.header_fields = ['table']
        importer.fields_description = Model.fields_description(
            fields=['name', 'table'])
        Model.insert(name='Model.IO.Test', table='io_test')
        importer.parse_row({'name': 'Model.IO.Test',
                            'table': 'io_test_other_table'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 1)

    def test_parse_row_with_unexting_multi_pks(self):
        Column = self.registry.System.Column
        importer = self.create_csv_importer(model='Model.System.Column')
        importer.header_pks = ['name', 'model']
        importer.header_fields = ['nullable']
        importer.fields_description = Column.fields_description(
            fields=['name', 'model', 'nullable'])
        importer.parse_row({'name': 'test',
                            'model': 'Model.System.Model',
                            'nullable': True})
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_parse_row_with_mapping_pks(self):
        Model = self.registry.System.Model
        importer = self.create_csv_importer(model='Model.System.Model')
        importer.header_external_id = 'name/EXTERNAL_ID'
        importer.header_fields = ['table']
        importer.fields_description = Model.fields_description(
            fields=['name', 'table'])
        model = Model.insert(name='Model.IO.Test', table='io_test')
        self.registry.IO.Mapping.set('import_mapping', model)
        importer.parse_row({'name/EXTERNAL_ID': 'import_mapping',
                            'table': 'io_test_other_table'})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 1)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(importer.updated_entries[0].table,
                         'io_test_other_table')

    def test_parse_row_with_unexisting_mapping_pks(self):
        Exporter = self.registry.IO.Exporter
        importer = self.create_csv_importer(model='Model.IO.Exporter')
        importer.header_external_id = 'id/EXTERNAL_ID'
        importer.header_fields = ['model', 'mode']
        importer.fields_description = Exporter.fields_description(
            fields=['id', 'model', 'mode'])
        importer.parse_row({'id/EXTERNAL_ID': 'import_mapping',
                            'model': 'Model.IO.Exporter',
                            'mode': 'Model.IO.Exporter.CSV'})
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)
        exporter = self.registry.IO.Mapping.get('Model.IO.Exporter',
                                                'import_mapping')
        self.assertEqual(exporter.mode, 'Model.IO.Exporter.CSV')
        self.assertEqual(exporter.model, 'Model.IO.Exporter')

    def test_parse_row_with_mapping(self):
        Model = self.registry.System.Model
        model = Model.insert(name='Model.IO.Test', table='io_test')
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Exporter')
        exporter = self.registry.IO.Exporter.CSV.insert(
            model='Model.IO.Exporter')
        importer.header_pks = ['id']
        importer.header_external_ids = {'model/EXTERNAL_ID': 'model'}
        importer.header_fields = ['mode']
        importer.fields_description = Importer.fields_description(
            fields=['id', 'model', 'mode'])
        self.registry.IO.Mapping.set('import_mapping', model)
        importer.parse_row({'id': str(exporter.id),
                            'model/EXTERNAL_ID': 'import_mapping',
                            'mode': "Model.IO.Exporter.CSV"})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 1)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(importer.updated_entries[0].model, 'Model.IO.Test')

    def test_parse_row_with_unexisting_mapping(self):
        Importer = self.registry.IO.Importer
        importer = self.create_csv_importer(model='Model.IO.Importer')
        importer.header_pks = ['id']
        importer.header_external_ids = {'model/EXTERNAL_ID': 'model'}
        importer.header_fields = ['mode']
        importer.fields_description = Importer.fields_description(
            fields=['id', 'model', 'mode'])
        importer.parse_row({'id': str(importer.importer.id),
                            'model/EXTERNAL_ID': 'import_mapping',
                            'mode': "Model.IO.Importer.CSV"})
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 1)

    def test_run(self):
        importer = self.create_csv_importer(
            model='Model.IO.Exporter',
            file_to_import=self.get_exporter_file_to_import())
        importer.run()
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_run_raise_at_end(self):
        importer = self.create_csv_importer()
        with self.assertRaises(CSVImporterException):
            importer.run()

        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 1)

    def test_run_raise_ignored(self):
        importer = self.create_csv_importer(csv_on_error='ignore')
        importer.run()
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 1)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from pkg_resources import iter_entry_points
from logging import getLogger

logger = getLogger(__name__)


def get_model_plugins(registry):
    res = []
    for i in iter_entry_points('anyblok.model.plugin'):
        logger.info('AnyBlok Load model plugin: %r' % i)
        res.append(i.load()(registry))

    return res


class ModelPluginBase:

    def __init__(self, registry):
        self.registry = registry

    def initialisation_tranformation_properties(self, properties,
                                                transformation_properties):
        """ Initialise the transform properties

        :param properties: the properties declared in the model
        :param new_type_properties: param to add in a new base if need
        """

    def declare_field(self, name, field, namespace, properties,
                      transformation_properties):
        """Declare a field in the model

        :param name: field name
        :param field: field instance
        :param namespace: the namespace of the model
        :param properties: the properties of the model
        :param transformation_properties: the transformation properties
        """

    def transform_base_attribute(self, attr, method, namespace, base,
                                 transformation_properties,
                                 new_type_properties):
        """ transform the attribute for the final Model

        :param attr: attribute name
        :param method: method pointer of the attribute
        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """

    def transform_base(self, namespace, base,
                       transformation_properties,
                       new_type_properties):
        """ transform the base for the final Model

        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """

    def insert_in_bases(self, new_base, namespace, properties,
                        transformation_properties):
        """Insert in a base the overload

        :param new_base: the base to be put on front of all bases
        :param namespace: the namespace of the model
        :param properties: the properties declared in the model
        :param transformation_properties: the properties of the model
        """

    def after_model_construction(self, base, namespace,
                                 transformation_properties):
        """Do some action with the constructed Model

        :param base: the Model class
        :param namespace: the namespace of the model
        :param transformation_properties: the properties of the model
        """
